/*
 * Copyright (C) 2009 Google, Inc.
 * Copyright (C) 2009 HTC Corporation
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/fs.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/uaccess.h>
#include <linux/msm_audio.h>

#include <mach/msm_qdsp6_audio.h>
#include <mach/debug_mm.h>

#define BUFSZ (0)

#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
#define AUDIOCTL_MSG_INFO
/* #undef AUDIOCTL_MSG_INFO */

#ifdef AUDIOCTL_MSG_INFO
#define TRACE(x...) pr_info("AUDIOCTL: "x)
#else
#define TRACE(x...) do{}while(0)
#endif
#endif CONFIG_KYOCERA_ORIGINAL_FEATURE

static DEFINE_MUTEX(voice_lock);
static int voice_started;

static struct audio_client *voc_tx_clnt;
static struct audio_client *voc_rx_clnt;

static int q6_voice_start(void)
{
	int rc = 0;

	mutex_lock(&voice_lock);

	if (voice_started) {
		pr_err("[%s:%s] busy\n", __MM_FILE__, __func__);
		rc = -EBUSY;
		goto done;
	}

	voc_tx_clnt = q6voice_open(AUDIO_FLAG_WRITE);
	if (!voc_tx_clnt) {
		pr_err("[%s:%s] open voice tx failed.\n", __MM_FILE__,
				__func__);
		rc = -ENOMEM;
		goto done;
	}

	voc_rx_clnt = q6voice_open(AUDIO_FLAG_READ);
	if (!voc_rx_clnt) {
		pr_err("[%s:%s] open voice rx failed.\n", __MM_FILE__,
				__func__);
		q6voice_close(voc_tx_clnt);
		rc = -ENOMEM;
	}

	voice_started = 1;
done:
	mutex_unlock(&voice_lock);
	return rc;
}

static int q6_voice_stop(void)
{
	mutex_lock(&voice_lock);
	if (voice_started) {
		q6voice_close(voc_tx_clnt);
		q6voice_close(voc_rx_clnt);
		voice_started = 0;
	}
	mutex_unlock(&voice_lock);
	return 0;
}

static int q6_open(struct inode *inode, struct file *file)
{
#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
    TRACE("open()\n");
#endif /* CONFIG_KYOCERA_ORIGINAL_FEATURE */
	return 0;
}

static int q6_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg)
{
	int rc;
	uint32_t n;
	uint32_t id[2];
	uint32_t mute_status;

	switch (cmd) {
	case AUDIO_SWITCH_DEVICE:
		rc = copy_from_user(&id, (void *)arg, sizeof(id));
#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
        TRACE("ioctl(AUDIO_SWITCH_DEVICE) adsp_id=0x%x acdb_id=0x%x\n", id[0], id[1]);
#endif /* CONFIG_KYOCERA_ORIGINAL_FEATURE */
		if (!rc)
			rc = q6audio_do_routing(id[0], id[1]);
		break;
	case AUDIO_SET_VOLUME:
		rc = copy_from_user(&n, (void *)arg, sizeof(n));
#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
        TRACE("ioctl(AUDIO_SET_VOLUME) volume=%d\n", n);
#endif /* CONFIG_KYOCERA_ORIGINAL_FEATURE */
		if (!rc)
			rc = q6audio_set_rx_volume(n);
		break;
	case AUDIO_SET_MUTE:
		rc = copy_from_user(&n, (void *)arg, sizeof(n));
#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
        TRACE("ioctl(AUDIO_SET_MUTE) mute=%d\n", n);
#endif /* CONFIG_KYOCERA_ORIGINAL_FEATURE */
		if (!rc) {
			if (voice_started) {
				if (n == 1)
					mute_status = STREAM_MUTE;
				else
					mute_status = STREAM_UNMUTE;
			} else {
				if (n == 1)
					mute_status = DEVICE_MUTE;
				else
					mute_status = DEVICE_UNMUTE;
			}

			rc = q6audio_set_tx_mute(mute_status);
		}
		break;
	case AUDIO_UPDATE_ACDB:
		rc = copy_from_user(&id, (void *)arg, sizeof(id));
		if (!rc)
			rc = q6audio_update_acdb(id[0], 0);
		break;
	case AUDIO_START_VOICE:
#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
        TRACE("ioctl(AUDIO_START_VOICE)\n");
#endif /* CONFIG_KYOCERA_ORIGINAL_FEATURE */
		rc = q6_voice_start();
		break;
	case AUDIO_STOP_VOICE:
#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
        TRACE("ioctl(AUDIO_STOP_VOICE)\n");
#endif /* CONFIG_KYOCERA_ORIGINAL_FEATURE */
		rc = q6_voice_stop();
		break;
	case AUDIO_REINIT_ACDB:
		rc = 0;
		break;
#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
  case AUDIO_SET_RX_MUTE:
    rc = copy_from_user(&n, (void *)arg, sizeof(n));
    if (!rc)
      rc = q6audio_set_rx_mute(n);
    break;

  case AUDIO_SET_SPEAKER_AMP:
    rc = copy_from_user(&n, (void *)arg, sizeof(n));
    if (!rc)
      rc = q6audio_set_speaker_amp(n);
    break;

    case AUDIO_BAM_CTL:
        rc = copy_from_user(&id, (void *)arg, sizeof(id));
        if (!rc)
            rc = q6audio_bam_ctl(id[0], id[1]);
        break;

    case AUDIO_GET_DEBUG_PARAM:
    {
        struct msm_audio_debug_param debug_param;
        rc = q6audio_get_debug_param((void*)&debug_param);
        if(!rc)
        {
            rc = copy_to_user((void*) arg, &debug_param, sizeof(debug_param));
        }
        break;
    }
#endif /* CONFIG_KYOCERA_ORIGINAL_FEATURE */
	default:
		rc = -EINVAL;
	}

	return rc;
}


static int q6_release(struct inode *inode, struct file *file)
{
#ifdef CONFIG_KYOCERA_ORIGINAL_FEATURE
    TRACE("release()\n");
#endif /* CONFIG_KYOCERA_ORIGINAL_FEATURE */
	return 0;
}

static struct file_operations q6_dev_fops = {
	.owner		= THIS_MODULE,
	.open		= q6_open,
	.ioctl		= q6_ioctl,
	.release	= q6_release,
};

struct miscdevice q6_control_device = {
	.minor	= MISC_DYNAMIC_MINOR,
	.name	= "msm_audio_ctl",
	.fops	= &q6_dev_fops,
};


static int __init q6_audio_ctl_init(void) {
	return misc_register(&q6_control_device);
}

device_initcall(q6_audio_ctl_init);
