/*
 * Copyright (C) 2010 Kionix, Inc.
 *      Written by Chris Hudson <chudson@kionix.com>
 * Copyright (c) 2010 Yamaha Corporation
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 */

#include <linux/err.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/i2c.h>
#include <linux/input.h>
#include <linux/uaccess.h>
#include <linux/workqueue.h>
#include <linux/irq.h>
#include <linux/gpio.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <mach/proc_comm_kyocera.h>


#define KXUD9_VERSION "1.0.0"
#define KXUD9_NAME    "kxud9"

/* for debugging */
#define DEBUG 0
#define DEBUG_DELAY 0
#define TRACE_FUNC() pr_debug(KXUD9_NAME ": <trace> %s()\n", __FUNCTION__)

/*
 * Default parameters
 */
#define KXUD9_DEFAULT_DELAY               100
#define KXUD9_MAX_DELAY                   2000

/*
 * Registers
 */
#define KXUD9_ACC_REG                     0x00

#define KXUD9_RESET_WRITE_REG             0x0a
#define KXUD9_RESET_WRITE_KEY             0xca

#define KXUD9_POWER_CONTROL_REG           0x0d
#define KXUD9_POWER_CONTROL_MASK          0x40
#define KXUD9_POWER_CONTROL_SHIFT         6

#define KXUD9_RANGE_REG                   0x0c
#define KXUD9_RANGE_MASK                  0x03
#define KXUD9_RANGE_SHIFT                 0
#define KXUD9_RANGE_8G                    0
#define KXUD9_RANGE_6G                    1
#define KXUD9_RANGE_4G                    2
#define KXUD9_RANGE_2G                    3

#define KXUD9_BANDWIDTH_REG               0x0c
#define KXUD9_BANDWIDTH_MASK              0xe0
#define KXUD9_BANDWIDTH_SHIFT             5
#define KXUD9_BANDWIDTH_2000HZ            3
#define KXUD9_BANDWIDTH_1000HZ            4
#define KXUD9_BANDWIDTH_500HZ             5
#define KXUD9_BANDWIDTH_100HZ             6
#define KXUD9_BANDWIDTH_50HZ              7

/* ABS axes parameter range [um/s^2] (for input event) */
#define GRAVITY_EARTH_RES                 11974
#define GRAVITY_EARTH                     9806550
#define ABSMIN_2G                         (-GRAVITY_EARTH * 2)
#define ABSMAX_2G                         (GRAVITY_EARTH * 2)

#if DEBUG
#define XOUT_H                            0x00
#define XOUT_L                            0x01
#define YOUT_H                            0x02
#define YOUT_L                            0x03
#define ZOUT_H                            0x04
#define ZOUT_L                            0x05
#define AUXOUT_H                          0x06
#define AUXOUT_L                          0x07
#define CTRL_REGC                         0x0c
#define CTRL_REGB                         0x0d
#define CTRL_REGA                         0x0e
#endif /* DEBUG */

struct acceleration {
    int x;
    int y;
    int z;
};

/*
 * Output data rate
 */
struct kxud9_odr {
        unsigned long delay;            /* min delay (msec) in the range of ODR */
        u8 odr;                         /* bandwidth register value */
};

static const struct kxud9_odr kxud9_odr_table[] = {
    {0,     KXUD9_BANDWIDTH_2000HZ},
    {1,     KXUD9_BANDWIDTH_1000HZ},
    {2,     KXUD9_BANDWIDTH_500HZ},
    {10,    KXUD9_BANDWIDTH_100HZ},
    {200,   KXUD9_BANDWIDTH_50HZ},
};

/*
 * Transformation matrix for chip mounting position
 */
static const int kxud9_position_map[][3][3] = {
    {{ 1,  0,  0}, { 0,  1,  0}, { 0,  0,  1}}, /* top/upper-left */
    {{ 0,  1,  0}, {-1,  0,  0}, { 0,  0,  1}}, /* top/upper-right */
    {{-1,  0,  0}, { 0, -1,  0}, { 0,  0,  1}}, /* top/lower-right */
    {{ 0, -1,  0}, { 1,  0,  0}, { 0,  0,  1}}, /* top/lower-left */
    {{-1,  0,  0}, { 0,  1,  0}, { 0,  0, -1}}, /* bottom/upper-left */
    {{ 0, -1,  0}, {-1,  0,  0}, { 0,  0, -1}}, /* bottom/upper-right */
    {{ 1,  0,  0}, { 0, -1,  0}, { 0,  0, -1}}, /* bottom/lower-right */
    {{ 0,  1,  0}, { 1,  0,  0}, { 0,  0, -1}}, /* bottom/lower-left */
};

/*
 * driver private data
 */
struct kxud9_data {
    atomic_t enable;                /* attribute value */
    atomic_t delay;                 /* attribute value */
    atomic_t position;              /* attribute value */
    struct acceleration last;       /* last measured data */
    struct mutex enable_mutex;
    struct mutex data_mutex;
    struct i2c_client *client;
    struct input_dev *input;
    struct delayed_work work;
#if DEBUG
    int suspend;
#endif
};

static int reg_val[3];
static int calib_val[3];
struct mutex regval_mutex;
struct mutex calib_mutex;


#define delay_to_jiffies(d) ((d)?msecs_to_jiffies(d):1)
#define actual_delay(d)     (jiffies_to_msecs(delay_to_jiffies(d)))

/* register access functions */
#define kxud9_read_bits(p,r) \
    ((i2c_smbus_read_byte_data((p)->client, r##_REG) & r##_MASK) >> r##_SHIFT)
#define kxud9_update_bits(p,r,v) \
    i2c_smbus_write_byte_data((p)->client, r##_REG, \
                              ((i2c_smbus_read_byte_data((p)->client,r##_REG) & ~r##_MASK) | ((v) << r##_SHIFT)))

static int debug_log_enable = 0;
#define KXUD9_CTRL_REG_ADDR     0x0c
#ifdef SENSOR_DPRINTK
#undef SENSOR_DPRINTK
#endif
#define SENSOR_DPRINTK(msg,...) if(debug_log_enable) printk("[KXUD9]:%d " msg,__LINE__,##__VA_ARGS__);


/*
 * Device dependant operations
 */
static int kxud9_power_up(struct kxud9_data *kxud9)
{
    kxud9_update_bits(kxud9, KXUD9_POWER_CONTROL, 1);

    return 0;
}

static int kxud9_power_down(struct kxud9_data *kxud9)
{
    kxud9_update_bits(kxud9, KXUD9_POWER_CONTROL, 0);

    return 0;
}

static int kxud9_hw_init(struct kxud9_data *kxud9)
{
    /* software reset */
    i2c_smbus_write_byte_data(kxud9->client, KXUD9_RESET_WRITE_REG, KXUD9_RESET_WRITE_KEY);
    msleep(1);

    /* rang setting */
    kxud9_update_bits(kxud9, KXUD9_RANGE, KXUD9_RANGE_2G);

    /* power down */
    kxud9_power_down(kxud9);

    return 0;
}

static int kxud9_get_enable(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxud9_data *kxud9 = i2c_get_clientdata(client);

    return atomic_read(&kxud9->enable);
}

static void kxud9_set_enable(struct device *dev, int enable)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxud9_data *kxud9 = i2c_get_clientdata(client);
    int delay = atomic_read(&kxud9->delay);

    mutex_lock(&kxud9->enable_mutex);

    if (enable) {                   /* enable if state will be changed */
        if (!atomic_cmpxchg(&kxud9->enable, 0, 1)) {
            SENSOR_DPRINTK("DEBUG:Set Enable\n");

            kxud9_power_up(kxud9);
            schedule_delayed_work(&kxud9->work, delay_to_jiffies(delay) + 1);
        }
    } else {                        /* disable if state will be changed */
        if (atomic_cmpxchg(&kxud9->enable, 1, 0)) {
            SENSOR_DPRINTK("DEBUG:Set Disable\n");

            cancel_delayed_work_sync(&kxud9->work);
            kxud9_power_down(kxud9);
        }
    }
    atomic_set(&kxud9->enable, enable);

    mutex_unlock(&kxud9->enable_mutex);
}

static int kxud9_get_delay(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxud9_data *kxud9 = i2c_get_clientdata(client);

    return atomic_read(&kxud9->delay);
}

static void kxud9_set_delay(struct device *dev, int delay)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxud9_data *kxud9 = i2c_get_clientdata(client);
    u8 odr;
    int i;

    /* determine optimum ODR */
    for (i = 1; (i < ARRAY_SIZE(kxud9_odr_table)) &&
             (actual_delay(delay) >= kxud9_odr_table[i].delay); i++)
        ;
    odr = kxud9_odr_table[i-1].odr;
    atomic_set(&kxud9->delay, delay);

    mutex_lock(&kxud9->enable_mutex);

    if (kxud9_get_enable(dev)) {
        cancel_delayed_work_sync(&kxud9->work);
        kxud9_update_bits(kxud9, KXUD9_BANDWIDTH, odr);
        schedule_delayed_work(&kxud9->work, delay_to_jiffies(delay) + 1);
    } else {
        kxud9_power_up(kxud9);
        kxud9_update_bits(kxud9, KXUD9_BANDWIDTH, odr);
        kxud9_power_down(kxud9);
    }

    mutex_unlock(&kxud9->enable_mutex);
}

static int kxud9_get_position(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxud9_data *kxud9 = i2c_get_clientdata(client);

    return atomic_read(&kxud9->position);
}

static void kxud9_set_position(struct device *dev, int position)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxud9_data *kxud9 = i2c_get_clientdata(client);

    atomic_set(&kxud9->position, position);
}

static int kxud9_measure(struct kxud9_data *kxud9, struct acceleration *accel)
{
    struct i2c_client *client = kxud9->client;
    u8 buf[6];
    int raw[3], data[3];
    int pos = atomic_read(&kxud9->position);
    long long g;
    int i, j;
#if DEBUG_DELAY
    struct timespec t;
#endif

#if DEBUG_DELAY
    getnstimeofday(&t);
#endif

    /* read acceleration data */
    if (i2c_smbus_read_i2c_block_data(client, KXUD9_ACC_REG, 6, buf) != 6) {
        dev_err(&client->dev,
                "I2C block read error: addr=0x%02x, len=%d\n",
                KXUD9_ACC_REG, 6);

        mutex_lock(&regval_mutex);

        reg_val[0] = 0;
        reg_val[1] = 0;
        reg_val[2] = 0;

        mutex_unlock(&regval_mutex);

        raw[0] = 0;
        raw[1] = 0;
        raw[2] = 0;


        SENSOR_DPRINTK("ERR:Register Read Error (Set All 0)\n");
    } else {
        mutex_lock(&regval_mutex);

        reg_val[0] = (int) (((buf[0] << 8) | buf[1]) >> 4);
        reg_val[1] = (int) (((buf[2] << 8) | buf[3]) >> 4);
        reg_val[2] = (int) (((buf[4] << 8) | buf[5]) >> 4);

        SENSOR_DPRINTK("DEBUG:Read Data x:0x%04x, y:0x%04x, z:0x%04x\n",reg_val[0],reg_val[1],reg_val[2]);

        mutex_lock(&calib_mutex);

        raw[0] = reg_val[0] - 2048 + calib_val[0];
        raw[1] = reg_val[1] - 2048 + calib_val[1];
        raw[2] = reg_val[2] - 2048 + calib_val[2];

        mutex_unlock(&calib_mutex);
        mutex_unlock(&regval_mutex);
    }

    /* for X, Y, Z axis */
    for (i = 0; i < 3; i++) {
        /* coordinate transformation */
        data[i] = 0;
        for (j = 0; j < 3; j++) {
            data[i] += raw[j] * kxud9_position_map[pos][i][j];
        }
        /* normalization */
        g = (long long) data[i] * GRAVITY_EARTH_RES;
        data[i] = g;
    }

    dev_dbg(&client->dev, "raw(%5d,%5d,%5d) => norm(%8d,%8d,%8d)\n",
            raw[0], raw[1], raw[2], data[0], data[1], data[2]);

#if DEBUG_DELAY
    dev_info(&client->dev, "%ld.%lds:raw(%5d,%5d,%5d) => norm(%8d,%8d,%8d)\n", t.tv_sec, t.tv_nsec,
             raw[0], raw[1], raw[2], data[0], data[1], data[2]);
#endif

    accel->x = data[0];
    accel->y = data[1];
    accel->z = data[2];

    return 0;
}

static void kxud9_work_func(struct work_struct *work)
{
    struct kxud9_data *kxud9 = container_of((struct delayed_work *)work, struct kxud9_data, work);
    struct acceleration accel;
    unsigned long delay = delay_to_jiffies(atomic_read(&kxud9->delay));

    kxud9_measure(kxud9, &accel);

    input_report_abs(kxud9->input, ABS_X, accel.x);
    input_report_abs(kxud9->input, ABS_Y, accel.y);
    input_report_abs(kxud9->input, ABS_Z, accel.z);
    input_sync(kxud9->input);

    SENSOR_DPRINTK("DEBUG:report Data x:%d,y:%d,z:%d\n",accel.x,accel.y,accel.z);

    mutex_lock(&kxud9->data_mutex);
    kxud9->last = accel;
    mutex_unlock(&kxud9->data_mutex);

    schedule_delayed_work(&kxud9->work, delay);
}

/*
 * Input device interface
 */
static int kxud9_input_init(struct kxud9_data *kxud9)
{
    struct input_dev *dev;
    int err;

    dev = input_allocate_device();
    if (!dev) {
        printk("[KXUD9] ERR:function error input_allocate_device()\n");
        return -ENOMEM;
    }
    dev->name = "accelerometer";
    dev->id.bustype = BUS_I2C;

    input_set_capability(dev, EV_ABS, ABS_MISC);
    input_set_abs_params(dev, ABS_X, ABSMIN_2G, ABSMAX_2G, 0, 0);
    input_set_abs_params(dev, ABS_Y, ABSMIN_2G, ABSMAX_2G, 0, 0);
    input_set_abs_params(dev, ABS_Z, ABSMIN_2G, ABSMAX_2G, 0, 0);
    input_set_drvdata(dev, kxud9);

    err = input_register_device(dev);
    if (err < 0) {
        input_free_device(dev);
        printk("[KXUD9] ERR:function error input_register_device()\n");

        return err;
    }
    kxud9->input = dev;

    return 0;
}

static void kxud9_input_fini(struct kxud9_data *kxud9)
{
    struct input_dev *dev = kxud9->input;

    input_unregister_device(dev);
    input_free_device(dev);
}

/*
 * sysfs device attributes
 */
static ssize_t kxud9_enable_show(struct device *dev,
                                 struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%d\n", kxud9_get_enable(dev));
}

static ssize_t kxud9_enable_store(struct device *dev,
                                  struct device_attribute *attr,
                                  const char *buf, size_t count)
{
    unsigned long enable = simple_strtoul(buf, NULL, 10);

    if ((enable == 0) || (enable == 1)) {
        kxud9_set_enable(dev, enable);
    }

    return count;
}

static ssize_t kxud9_delay_show(struct device *dev,
                                struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%d\n", kxud9_get_delay(dev));
}

static ssize_t kxud9_delay_store(struct device *dev,
                                 struct device_attribute *attr,
                                 const char *buf, size_t count)
{
    unsigned long delay = simple_strtoul(buf, NULL, 10);

    if (delay > KXUD9_MAX_DELAY) {
        delay = KXUD9_MAX_DELAY;
    }

    kxud9_set_delay(dev, delay);

    return count;
}

static ssize_t kxud9_position_show(struct device *dev,
                                   struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%d\n", kxud9_get_position(dev));
}

static ssize_t kxud9_position_store(struct device *dev,
                                    struct device_attribute *attr,
                                    const char *buf, size_t count)
{
    unsigned long position;

    position = simple_strtoul(buf, NULL,10);
    if ((position >= 0) && (position <= 7)) {
        kxud9_set_position(dev, position);
    }

    return count;
}

static ssize_t kxud9_wake_store(struct device *dev,
                                struct device_attribute *attr,
                                const char *buf, size_t count)
{
    struct input_dev *input = to_input_dev(dev);
    static atomic_t serial = ATOMIC_INIT(0);

    input_report_abs(input, ABS_MISC, atomic_inc_return(&serial));

    return count;
}

static ssize_t kxud9_data_show(struct device *dev,
                               struct device_attribute *attr, char *buf)
{
    struct input_dev *input = to_input_dev(dev);
    struct kxud9_data *kxud9 = input_get_drvdata(input);
    struct acceleration accel;

    mutex_lock(&kxud9->data_mutex);
    accel = kxud9->last;
    mutex_unlock(&kxud9->data_mutex);

    return sprintf(buf, "%d %d %d\n", accel.x, accel.y, accel.z);
}

#if DEBUG
static ssize_t kxud9_debug_reg_show(struct device *dev,
                                     struct device_attribute *attr, char *buf)
{
    struct input_dev *input = to_input_dev(dev);
    struct kxud9_data *kxud9 = input_get_drvdata(input);
    struct i2c_client *client = kxud9->client;
    ssize_t count = 0;
    u8 reg;

    reg = i2c_smbus_read_byte_data(client, XOUT_H);
    count += sprintf(&buf[count], "%02x: %d\n", XOUT_H, reg);
    reg = i2c_smbus_read_byte_data(client, XOUT_L);
    count += sprintf(&buf[count], "%02x: %d\n", XOUT_L, reg);
    reg = i2c_smbus_read_byte_data(client, YOUT_H);
    count += sprintf(&buf[count], "%02x: %d\n", YOUT_H, reg);
    reg = i2c_smbus_read_byte_data(client, YOUT_L);
    count += sprintf(&buf[count], "%02x: %d\n", YOUT_L, reg);
    reg = i2c_smbus_read_byte_data(client, ZOUT_H);
    count += sprintf(&buf[count], "%02x: %d\n", ZOUT_H, reg);
    reg = i2c_smbus_read_byte_data(client, ZOUT_L);
    count += sprintf(&buf[count], "%02x: %d\n", ZOUT_L, reg);
    reg = i2c_smbus_read_byte_data(client, AUXOUT_H);
    count += sprintf(&buf[count], "%02x: %d\n", AUXOUT_H, reg);
    reg = i2c_smbus_read_byte_data(client, AUXOUT_L);
    count += sprintf(&buf[count], "%02x: %d\n", AUXOUT_L, reg);
    reg = i2c_smbus_read_byte_data(client, CTRL_REGC);
    count += sprintf(&buf[count], "%02x: %d\n", CTRL_REGC, reg);
    reg = i2c_smbus_read_byte_data(client, CTRL_REGB);
    count += sprintf(&buf[count], "%02x: %d\n", CTRL_REGB, reg);
    reg = i2c_smbus_read_byte_data(client, CTRL_REGA);
    count += sprintf(&buf[count], "%02x: %d\n", CTRL_REGA, reg);

    return count;
}

static int kxud9_suspend(struct i2c_client *client, pm_message_t mesg);
static int kxud9_resume(struct i2c_client *client);

static ssize_t kxud9_debug_suspend_show(struct device *dev,
                                         struct device_attribute *attr, char *buf)
{
    struct input_dev *input = to_input_dev(dev);
    struct kxud9_data *kxud9 = input_get_drvdata(input);

    return sprintf(buf, "%d\n", kxud9->suspend);
}

static ssize_t kxud9_debug_suspend_store(struct device *dev,
                                          struct device_attribute *attr,
                                          const char *buf, size_t count)
{
    struct input_dev *input = to_input_dev(dev);
    struct kxud9_data *kxud9 = input_get_drvdata(input);
    struct i2c_client *client = kxud9->client;
    unsigned long suspend = simple_strtoul(buf, NULL, 10);

    if (suspend) {
        pm_message_t msg;
        kxud9_suspend(client, msg);
    } else {
        kxud9_resume(client);
    }

    return count;
}
#endif /* DEBUG */


static ssize_t kxud9_regdata_show(struct device *dev,
                                  struct device_attribute *attr, char *buf)
{
    ssize_t ret;

    mutex_lock(&regval_mutex);

    ret = sprintf(buf, "%d %d %d\n", reg_val[0], reg_val[1], reg_val[2]);

    mutex_unlock(&regval_mutex);

    return ret;

}
static ssize_t kxud9_calib_store(struct device *dev,
                                 struct device_attribute *attr,
                                 const char *buf, size_t count)
{
    int i,j,ptr,minus;


    for (i = 0, j = 0, ptr = 0, minus = 0; i < count; i++) {
        switch (buf[i]) {
            case 0x2d:
                minus = 1;
                ptr++;
                break;

            case 0x20:
                mutex_lock(&calib_mutex);

                calib_val[j] = simple_strtoul(&buf[ptr], NULL, 10);
                if (minus) {
                    calib_val[j] *= -1;
                }

                mutex_unlock(&calib_mutex);

                j++;
                ptr = i + 1;
                minus = 0;
                break;

            case 0x0a:
                mutex_lock(&calib_mutex);

                calib_val[j] = simple_strtoul(&buf[ptr], NULL, 10);
                if (minus) {
                    calib_val[j] *= -1;
                }

                mutex_unlock(&calib_mutex);

                break;

            default:
                break;
        }
    }

    printk("[KXUD9] DEBUG:Calib New Data x:%d, y:%d, z:%d\n",calib_val[0],calib_val[1],calib_val[2]);

    return count;
}


static ssize_t kxud9_dbglog_show(struct device *dev,
                                  struct device_attribute *attr, char *buf)
{
    unsigned char reg_dat[2];
    struct input_dev *input = to_input_dev(dev);
    struct kxud9_data *kxud9 = input_get_drvdata(input);
    struct i2c_client *client = kxud9->client;
    int ret;

    debug_log_enable ^= 1;

    if (debug_log_enable) {
        if (i2c_smbus_read_i2c_block_data(client, KXUD9_CTRL_REG_ADDR, 2, reg_dat) == 2) {
            ret = sprintf(buf, "[KXUD9] DEBUG:DEBUG ON(CTRL_REGC:0x%04x, CTRL_REGB:0x%04x)\n",reg_dat[0],reg_dat[1]);
            printk("[KXUD9] DEBUG:DEBUG ON(CTRL_REGC:0x%04x, CTRL_REGB:0x%04x)\n",reg_dat[0],reg_dat[1]);
        }
        else {
            ret = sprintf(buf, "[KXUD9] ERR:DEBUG ON(Control Register Read Error)\n");
            printk("[KXUD9] ERR:DEBUG ON(Control Register Read Error)\n");
        }
    }
    else {
        ret = sprintf(buf, "[KXUD9] DEBUG:DEBUG OFF\n");
        printk("[KXUD9] DEBUG:DEBUG OFF\n");
    }

    return ret;
}


static DEVICE_ATTR(enable, S_IRUGO|S_IWUSR|S_IWGRP,
                   kxud9_enable_show, kxud9_enable_store);
static DEVICE_ATTR(delay, S_IRUGO|S_IWUSR|S_IWGRP,
                   kxud9_delay_show, kxud9_delay_store);
static DEVICE_ATTR(position, S_IRUGO|S_IWUSR,
                   kxud9_position_show, kxud9_position_store);
static DEVICE_ATTR(wake, S_IWUSR|S_IWGRP, NULL, kxud9_wake_store);
static DEVICE_ATTR(data, S_IRUGO,
                   kxud9_data_show, NULL);

#if DEBUG
static DEVICE_ATTR(debug_reg, S_IRUGO,
                   kxud9_debug_reg_show, NULL);
static DEVICE_ATTR(debug_suspend, S_IRUGO|S_IWUSR,
                   kxud9_debug_suspend_show, kxud9_debug_suspend_store);
#endif /* DEBUG */

static DEVICE_ATTR(regdata, S_IRUGO, kxud9_regdata_show, NULL);
static DEVICE_ATTR(calib, S_IWUSR|S_IWGRP, NULL, kxud9_calib_store);
static DEVICE_ATTR(dbglog, S_IRUGO, kxud9_dbglog_show, NULL);


static struct attribute *kxud9_attributes[] = {
    &dev_attr_enable.attr,
    &dev_attr_delay.attr,
    &dev_attr_position.attr,
    &dev_attr_wake.attr,
    &dev_attr_data.attr,
#if DEBUG
    &dev_attr_debug_reg.attr,
    &dev_attr_debug_suspend.attr,
#endif /* DEBUG */
    &dev_attr_regdata.attr,
    &dev_attr_calib.attr,
    &dev_attr_dbglog.attr,
    NULL
};

static struct attribute_group kxud9_attribute_group = {
    .attrs = kxud9_attributes
};

/*
 * I2C client
 */
static int kxud9_detect(struct i2c_client *client, struct i2c_board_info *info)
{
    u8 r;

    /* software reset */
    i2c_smbus_write_byte_data(client, KXUD9_RESET_WRITE_REG, KXUD9_RESET_WRITE_KEY);
    msleep(1);

    r = i2c_smbus_read_byte_data(client, KXUD9_RANGE_REG);
    if (r != 0xe3)
        return -ENODEV;

    return 0;
}


static int kxud9_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
    struct kxud9_data *kxud9;
    int err;


    printk("[KXUD9] START:kxud9_probe()\n");

    /* setup private data */
    kxud9 = kzalloc(sizeof(struct kxud9_data), GFP_KERNEL);
    if (!kxud9) {
        printk("[KXUD9] ERR:function error input_allocate_device()\n");
        err = -ENOMEM;
        goto error_0;
    }

    mutex_init(&kxud9->enable_mutex);
    mutex_init(&kxud9->data_mutex);

    /* setup i2c client */
    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
        printk("[KXUD9] ERR:function error i2c_check_functionality()\n");
        err = -ENODEV;
        goto error_1;
    }
    i2c_set_clientdata(client, kxud9);
    kxud9->client = client;

    /* detect and init hardware */
    if ((err = kxud9_detect(client, NULL))) {
        printk("[KXUD9] ERR:function error kxud9_detect()\n");

        goto error_1;
    }
    dev_info(&client->dev, "%s found\n", id->name);

    kxud9_hw_init(kxud9);
    kxud9_set_delay(&client->dev, KXUD9_DEFAULT_DELAY);
    kxud9_set_position(&client->dev, CONFIG_INPUT_KXUD9_POSITION);

    /* setup driver interfaces */
    INIT_DELAYED_WORK(&kxud9->work, kxud9_work_func);

    err = kxud9_input_init(kxud9);
    if (err < 0) {
        printk("[KXUD9] ERR:function error kxud9_input_init()\n");

        goto error_1;
    }

    err = sysfs_create_group(&kxud9->input->dev.kobj, &kxud9_attribute_group);
    if (err < 0) {
        printk("[KXUD9] ERR:function error sysfs_create_group()\n");

        goto error_2;
    }

    reg_val[0]   = 0;
    reg_val[1]   = 0;
    reg_val[2]   = 0;
    calib_val[0] = 0;
    calib_val[1] = 0;
    calib_val[2] = 0;

    printk("[KXUD9] DEBUG:Calib Init Data x:%d, y:%d, z:%d\n",calib_val[0],calib_val[1],calib_val[2]);

    mutex_init(&regval_mutex);
    mutex_init(&calib_mutex);

    printk("[KXUD9] END:kxud9_probe()\n");

    return 0;

error_2:
    kxud9_input_fini(kxud9);
error_1:
    kfree(kxud9);
error_0:

    printk("[KXUD9] END:kxud9_probe()\n");

    return err;
}

static int kxud9_remove(struct i2c_client *client)
{
    struct kxud9_data *kxud9 = i2c_get_clientdata(client);

    printk("[KXUD9] START:kxud9_remove()\n");

    kxud9_set_enable(&client->dev, 0);

    sysfs_remove_group(&kxud9->input->dev.kobj, &kxud9_attribute_group);
    kxud9_input_fini(kxud9);

    kfree(kxud9);

    printk("[KXUD9] END:kxud9_remove()\n");

    return 0;
}

static int kxud9_suspend(struct i2c_client *client, pm_message_t mesg)
{
    struct kxud9_data *kxud9 = i2c_get_clientdata(client);

    printk("[KXUD9] START:kxud9_suspend()\n");

    mutex_lock(&kxud9->enable_mutex);

    if (kxud9_get_enable(&client->dev)) {
        cancel_delayed_work_sync(&kxud9->work);
        kxud9_power_down(kxud9);
    }

#if DEBUG
    kxud9->suspend = 1;
#endif

    mutex_unlock(&kxud9->enable_mutex);

    printk("[KXUD9] END:kxud9_suspend()\n");

    return 0;
}

static int kxud9_resume(struct i2c_client *client)
{
    struct kxud9_data *kxud9 = i2c_get_clientdata(client);
    int delay = atomic_read(&kxud9->delay);

    printk("[KXUD9] START:kxud9_resume()\n");

    kxud9_hw_init(kxud9);
    kxud9_set_delay(&client->dev, delay);

    mutex_lock(&kxud9->enable_mutex);

    if (kxud9_get_enable(&client->dev)) {
        kxud9_power_up(kxud9);
        schedule_delayed_work(&kxud9->work, delay_to_jiffies(delay) + 1);
    }

#if DEBUG
    kxud9->suspend = 0;
#endif

    mutex_unlock(&kxud9->enable_mutex);

    printk("[KXUD9] END:kxud9_resume()\n");

    return 0;
}

static const struct i2c_device_id kxud9_id[] = {
    {KXUD9_NAME, 0},
    {},
};

MODULE_DEVICE_TABLE(i2c, kxud9_id);

static struct i2c_driver kxud9_driver = {
    .driver = {
        .name = KXUD9_NAME,
        .owner = THIS_MODULE,
    },
    .probe = kxud9_probe,
    .remove = kxud9_remove,
    .suspend = kxud9_suspend,
    .resume = kxud9_resume,
    .id_table = kxud9_id,
};

static int __init kxud9_init(void)
{
    return i2c_add_driver(&kxud9_driver);
}

static void __exit kxud9_exit(void)
{
    i2c_del_driver(&kxud9_driver);
}

module_init(kxud9_init);
module_exit(kxud9_exit);

MODULE_DESCRIPTION("KXUD9 accelerometer driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(KXUD9_VERSION);
